import json
from typing import List

from santoku.aws.secretsmanager import SecretsManagerHandler

from slack import WebClient
from slack.errors import SlackApiError


class SlackBotHandler:
    """
    Class to manage the connection with an application bot of slack and the real time messaging to
    a chanel.

    Notes
    -----
    More information on slack apps: [1]

    References
    ----------
    [1] :
    https://api.slack.com/authentication/basics

    """

    def __init__(self, api_token: str):
        """
        Initialize the slack client.

        Parameters
        ----------
        api_token : str
            The api token of the application bot.

        Notes
        -----
            More information on the Slack Web API: [1].
            More information on the Slack Client: [2].

        References
        ----------
        [1] :
        https://slack.dev/python-slackclient/basic_usage.html

        [2]:
        https://github.com/slackapi/python-slackclient

        """
        self.client = WebClient(token=api_token)

    @classmethod
    def from_aws_secrets_manager(cls, secret_name: str, secret_key: str = "API_TOKEN"):
        """
        Retrieve the slack app oauth token from AWS Secrets Manager and initialize the slack client.
        Requires that AWS credentials with the appropriate permissions are located somewhere on the
        AWS credential chain in the local machine.

        Paramaters
        ----------
        secret_name : str
            Name or ARN for the secret containing the token needed for the Slack bot authentication.

        secret_key : str, optional
            Key of the stored secret. (By default "API_TOKEN" will be the key that stores the slack
            app oauth token.)

        See Also
        --------
        __init__ : this method calls the constructor.

        """
        secrets_manager = SecretsManagerHandler()
        credential_info = secrets_manager.get_secret_value(secret_name=secret_name)
        return cls(api_token=credential_info[secret_key])

    def send_message(self, channel: str, **kwargs) -> None:
        """
        Send a custom `message` to the `channel`.

        Parameters
        ----------
        channel : str
            To where the bot will send the message.
        message : str
            Custom message that will be sent to the chanel.
        blocks : List[Dict[str, Any]]
            JSON formatted string for a block structure [2]

        Raises
        ------
        SlackBotError
            If the authentication token is invalid or if the specified channel cannot be found.

        Notes
        -----
            In order to successfully send the message, the bot must be added to the workspace, it
            also needs to have the scopes to send message to the specified channel approved, this
            can be configured in: [1].

        References
        ----------
        [1] :
        https://wiris.slack.com/apps/manage

        [2] :
        https://api.slack.com/block-kit
        """
        self.client.chat_postMessage(channel=channel, **kwargs)

    def send_process_report(
        self,
        channel: str,
        process_name: str,
        messages: List[str],
        is_success: bool,
        context_message: str,
        context_url: str,
        context_img: str = "https://bitbucket-assetroot.s3.amazonaws.com/c/photos/2020/Aug/06/1924969399-7-lambda-moodle-referers-logo_avatar.png",
    ) -> None:
        """
        Posts a message to `channel`, using Slack's blocks API to

        Parameters
        ----------
        channel: str
            To where the bot will send the message.
        process_name: str
            ID of the process posting the report.
        messages: List[str]
            List of report messages to send to Slack
        is_success: bool
            Process success status
        context_url: str
            URL to include in the context
        context_message: str
            Message to display in the context
        context_img: str
            Image to include in the context thumbnail
        """

        blocks: List[dict] = [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": f"{process_name} {':heavy_check_mark:' if is_success else ':x:'}",
                    "emoji": True,
                },
            },
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": "\n".join([f"- {m}" for m in messages])},
            },
        ]

        if context_message and context_url and context_img:
            blocks.append(
                {
                    "type": "context",
                    "elements": [
                        {"type": "image", "image_url": context_img, "alt_text": "images"},
                        {"type": "mrkdwn", "text": f"<{context_url}|{context_message}>"},
                    ],
                }
            )

        self.send_message(channel=channel, blocks=blocks)



import json

import jsonschema

from jsonschema.exceptions import ValidationError, SchemaError
from typing import Any, Dict, List, Optional, Union


"""
Note: This module is a temporary standalone solution, subject to change. Pending implementation of
either a third party solution or a fully featured version someday.
"""


class ConfigurationError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class IllegalAccessPattern(Exception):
    def __init__(self) -> None:
        super().__init__(f"Acessing non-final nodes in the settings hierarchy is not allowed")


class SettingError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class SchemaViolation(Exception):
    def __init__(self) -> None:
        super().__init__("The settings do not conform to the specified schema.")


class ConfigurationManager:
    """
    The purpose of this abstract class is to centrally manage configuration parameters for an
    entire application.

    We define one such parameters as a `setting`. A `setting` is a key-value pair. We organize
    settings in nested collections (essentially JSON objects, except arrays cannot contain objects).

    A `configuration` is a named collection of settings. We represent a configuration by a JSON
    object following the structure of this example:

    {
        "name": "my_configuration_name",
        "settings": {
            "boolean_setting": true,
            "int_setting": 42,
            "object_setting": {
                "nested_float_setting": 3.14,
                "nested_string_setting": ""
            }
        }
    }

    where the value of "settings" is an arbitrary JSON object (the collection of settings).

    Optionally, a schema can be provided, using the JSON Schema specification, to which the settings
    object must conform. Here's how the schema for the above example would look like:

    {
        "type": "object",
        "properties": {
            "boolean_setting": {"type": "boolean"},
            "int_setting": {"type": "integer"},
            "object_setting": {
                "type": "object",
                "properties": {
                    "nested_float_setting": {"type": "number"},
                    "nested_string_setting": {"type": "string"}
                },
                "required": ["nested_float_setting", "nested_string_setting"],
                "additionalProperties": false
            }
        },
        "required": ["boolean_setting", "integer_setting", "object_setting"],
        "additionalProperties": false
    }

    The settings on each configuration must follow the schema if given.
    """

    JSON = Union[dict, list]  # naive (i.e. wrong) type hint for parsed JSON objects
    configuration_list_schema: dict = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "description": "A list of configurations to be parsed and managed by this class.",
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of the configuration."},
                "settings": {
                    "description": "Arbitrary object with (potentially nested) settings.",
                    "type": "object",
                },
            },
            "required": ["name", "settings"],
            "additionalProperties": False,
        },
    }

    def __init__(
        self,
        configurations: List[dict] = None,
        schema: dict = None,
        initial_configuration: str = None,
    ):
        """
        A list of available `configurations` and a `schema` can be passed to be stored in the
        manager. The `initial_configuration` is an optional value. It must be the name of one of
        the `configurations`, which will become the active configuration. This is simply to save
        the user the initial call to `apply_configuration`.

        Parameters
        ----------
        configurations : List[Dict[str, Union[str, Dict[str, Any]]]], optional
            Collection of configurations with different combination of settings.
        schema : Dict[str, Any], optional
            Structure of the settings with the corresponding types of each setting value.
        initial_configuration: str, optional
            Name of the configuration in the `configurations` that will be set as the initial one.

        Raises
        ------
        ConfigurationError
            If one or more elements in `configurations` does not contain the "name" or "settings" keys.

        Notes
        -----
        The schema must conform to the appropriate draft of the JSON Schema specification, given by
        the reserved parameter "$schema" [1].
        The list of configurations must conform to the following structure, also checked via JSON
        Schema:
            [
                {
                    "name": <name_of_the_configuration>,
                    "settings": {
                        <key>: <value>,
                        ...
                    }
                },
                ...
            ]

        See also
        --------
        check_schema
            Used to validate the schema itself against the JSON Schema specification.
        validate_schema
            Called to check whether each configuration follows the schema (if provided).
        define_configuration
            Called to add each configuration to the list of available configurations.
        apply_configuration
            Called to activate the `initial_configuration` (if provided).

        References
        ----------
        [1] :
        https://json-schema.org/understanding-json-schema/basics.html#declaring-a-json-schema
        """
        if schema:
            self.check_schema(schema=schema)
        self.schema = schema

        self.configurations = {}
        if configurations:
            jsonschema.validate(instance=configurations, schema=self.configuration_list_schema)
            for config in configurations:
                settings = config["settings"]
                self.validate_schema(settings=settings)
                self.configurations[config["name"]] = settings

        self.active_configuration: Optional[str] = None
        if initial_configuration:
            self.apply_configuration(name=initial_configuration)

    @classmethod
    def from_json(
        cls,
        configurations_file_path: str,
        schema_file_path: str = None,
        initial_configuration: str = None,
    ) -> "ConfigurationManager":
        """
        Initialize the manager with the configurations stored in a JSON file.

        Parameters
        ----------
        configurations_file_path: str
            Path to the file containing configurations to be loaded.
        schema_file_path: str
            Path to the schema for the settings. Optional.
        initial_configuration: str
            Name of the initial configuration. Optional.
        """
        with open(configurations_file_path, "r") as f:
            configurations = json.load(f)

        if schema_file_path:
            with open(schema_file_path, "r") as f:
                schema = json.load(f)
        else:
            schema = None

        return cls(
            configurations=configurations,
            schema=schema,
            initial_configuration=initial_configuration,
        )

    @staticmethod
    def check_schema(schema: dict) -> None:
        """
        Throws an exception if the schema is not a valid instance of the JSON Schema specification.
        The schema can (and should) specify which draft of JSON Schema it follows using the
        reserved keyword "$schema" [1]. Supports drafts 4, 6 and 7 of JSON Schema [2].
        If not specified, the check is performed against draft 7, the newest version at the time of
        writing this module.

        Parameters
        ----------
        schema : dict
            The schema to be checked.
        
        References
        ----------
        [1] :
        https://json-schema.org/understanding-json-schema/basics.html#declaring-a-json-schema
        [2] :
        https://json-schema.org/understanding-json-schema/reference/schema.html#schema
        """
        specification = schema.get("$schema", None)
        if specification == "http://json-schema.org/draft-07/schema#":
            jsonschema.Draft7Validator.check_schema(schema=schema)
        elif specification == "http://json-schema.org/draft-06/schema#":
            jsonschema.Draft6Validator.check_schema(schema=schema)
        elif specification == "http://json-schema.org/draft-04/schema#":
            jsonschema.Draft4Validator.check_schema(schema=schema)
        elif specification is None:
            # Default to draft 7 if unspecified
            jsonschema.Draft7Validator.check_schema(schema=schema)
        else:
            # Throw exception if specified but not supported
            raise SchemaError(f"JSON Schema specification ({specification}) is not supported.")

    def validate_schema(self, settings: JSON) -> None:
        """
        Throws an exception if a schema was specified and the settings do not comply with it.

        Parameters
        ----------
        settings : Union[dict, list]
            An arbitrary JSON object, parsed to use Python lists and dictionaries.

        Raises
        ------
        SchemaViolation
            If a schema exists and `settings` does not comply with it.
        """
        if self.schema:
            try:
                jsonschema.validate(instance=settings, schema=self.schema)
            except:
                raise SchemaViolation

    def define_configuration(self, name: str, settings: JSON, override: bool = False) -> None:
        """
        Adds a new configuration to the set of available configurations, which can be refrenced by
        name.

        If the configuration already exists, an exception will be thrown unless `override` is set to
        True.

        The given `configuration` must follow the given schema, if any.

        Parameters
        ----------
        name : str
            ID of the configuration.
        settings : Union[dict, list]
            An arbitrary JSON object, parsed to use dictionaries and lists.
        override : bool, optional
            If set to True, previous configurations with the same name will be overridden.

        Raises
        ------
        ConfigurationError
            If `override` is set to False and a configuration with the same name already exists.

        ConfigurationError
            If the configuration does not follow the schema.
        """
        if not override and name in self.configurations:
            raise ConfigurationError(f"Configuration '{name}' already exists.")

        self.validate_schema(settings=settings)
        self.configurations[name] = settings

        # If the modified config is the active one, reapply it
        if name == self.active_configuration:
            self.apply_configuration(name=name)

    def list_configurations(self) -> List[str]:
        """
        Lists all the available configurations by name.

        Returns
        -------
        List[str]
            A list of the names of the available configurations
        """
        return list(self.configurations.keys())

    def get_configuration(self, name: str) -> JSON:
        """
        Retrieves all the settings of an available configuration.

        Parameters
        ----------
        name : str
            Name of the configuration.

        Returns
        -------
        Union[dict, list]
            Arbitrary JSON-like object with the contents of the requested configuration.

        Raises
        ------
        ConfigurationError
            If no configuration with that name exists.
        """
        try:
            return self.configurations[name]
        except KeyError:
            raise ConfigurationError(f"Configuration '{name}' has not been defined")

    def apply_configuration(self, name: str) -> None:
        """
        Sets a given configuration as the active configuration.

        Parameters
        ----------
        name: str
            ID of the configuration.

        Raises
        ------
        ConfigurationError
            If no configuration with that name exists.
        """
        if name not in self.configurations:
            raise ConfigurationError(f"Configuration '{name}' has not been defined")
        self.active_configuration = name

    def get_active_configuration(self) -> JSON:
        """
        Retrieve all settings from the active configuration.

        Returns
        -------
        Union[dict, list]
            Arbitrary JSON-like object with the contents of the active configuration.

        Raises
        ------
        ConfigurationError
            If no configuration is set to active
        """
        if not self.active_configuration:
            raise ConfigurationError("No active configuration")
        return self.get_configuration(name=self.active_configuration)

    def get_setting(self, *args) -> Any:
        """
        Retrieve the value of a particular setting in the active configuration. To access a nested
        value, pass all necessary keys needed to traverse the nesting (up to a leaf node) in order.

        Parameters
        ----------
        args
            Sequence of keys needed to traverse the configuration tree up to a leaf node.

        Returns
        -------
        Any
            Value of the setting.

        Raises
        ------
        SettingError
            If a key that is not in the active configuration is given.
        IllegalAccessPattern
            If the keys do not reach a leaf node in the configuration tree.
        
        Example
        -------
        Given the following settings:
        {
            "my_setting": 3,
            "nested_settings": {"foo": "bar", "stuff": ["this", "that"]}
        }
        You can access each of the individual settings, but only final values:
        >>> configuration_manager.get_setting("my_setting")
        3
        >>> configuration_manager.get_setting("nested_settings", "foo")
        'bar'
        >>> configuration_manager.get_setting("nested_settings", "stuff")
        ['this', 'that']
        >>> configuration_manager.get_setting("nested_settings")
        IllegalAccessPattern: Acessing non-final nodes in the settings hierarchy is not allowed
        """
        setting = self.get_active_configuration()
        for key in args:
            if key not in setting:
                raise SettingError(f"Setting '{key}' not found amongst {list(setting)}.")
            setting = setting[key]

        if type(setting) == dict:
            raise IllegalAccessPattern

        return setting



from urllib.parse import ParseResult

import pytest
from santoku.utils.url_handler import URLHandler


@pytest.mark.parametrize(
    argnames=("input_subdomain", "expected_subdomain_list"),
    argvalues=(
        ("sub1.sub2", ["sub1", "sub2"]),
        ("sub1", ["sub1"]),
        ("sub1.", ["sub1", ""]),
        ("sub1..", ["sub1", "", ""]),
        ("", []),
        (None, []),
    ),
    scope="function",
)
def test_string_subdomain_is_processed_as_expected(input_subdomain, expected_subdomain_list):
    output_subdomain_list = URLHandler.process_subdomain(subdomain=input_subdomain)
    assert output_subdomain_list == expected_subdomain_list


@pytest.mark.parametrize(
    argnames=("input_url", "expected_components"),
    argvalues=[
        (
            "https://sub2.sub1.example.com/path?query#fragment",
            ("https", ["sub2", "sub1"], "example", "com", "/path"),
        ),
        (
            "https://sub1.example.com/path/",
            ("https", ["sub1"], "example", "com", "/path/"),
        ),
        ("http://sub2.sub1.example.es", ("http", ["sub2", "sub1"], "example", "es", None)),
        ("https://sub1.example.co.uk", ("https", ["sub1"], "example", "co.uk", None)),
        ("https://example.com", ("https", [], "example", "com", None)),
        ("https://example.XN--1QQW23A", ("https", [], "example", "XN--1QQW23A", None)),
        ("https://user@example.co.uk:8000", ("https", [], "example", "co.uk", None)),
        # Note that com.uk isn't a valid TLD (but it is a valid URL), so it is expected that the
        # function would consider 'example' as subdomain, 'com' as domain, and 'uk' as suffix
        ("https://example.com.uk", ("https", ["example"], "com", "uk", None)),
        ("www.example.com", (None, ["www"], "example", "com", None)),
        ("https://localhost.example", ("https", ["localhost"], "example", None, None)),
        ("https://sub1.localhost.example", ("https", ["sub1", "localhost"], "example", None, None)),
        (
            "https://sub1..localhost.example",
            ("https", ["sub1", "", "localhost"], "example", None, None),
        ),
        ("localhost", (None, [], "localhost", None, None)),
        ("https://125.0.0.0", ("https", [], "125.0.0.0", None, None)),
        ("com", (None, [], None, "com", None)),
        ("*", (None, [], "*", None, None)),
        ("//", (None, [], None, None, None)),
        ("", (None, [], None, None, None)),
        ("https:///integration/ckeditor", ("https", [], None, None, "/integration/ckeditor")),
    ],
    scope="function",
)
def test_url_are_split_into_components_as_expected(input_url, expected_components):
    output_components = URLHandler.split_url_into_components(url=input_url)
    assert output_components == expected_components


@pytest.mark.parametrize(
    argnames=("input_url", "expected_result"),
    argvalues=[
        ("https://sub2.sub1.example.com/path?query#fragment", True),
        ("https://sub2.sub1.example.com", True),
        ("https://example.com", True),
        ("https://example.co.uk", True),
        ("https://user@example.co.uk:8000", True),
        # Note that com.uk isn't a valid TLD (but it is a usable URL), so it is expected that the
        # funcition would consider 'example' as subdomain, 'com' as domain, and 'uk' as suffix
        ("https://example.com.uk", True),
        ("www.example.com", True),
        # Note that this URL is not usable because it doesn't contain suffix, so 'localhost' will be
        # be considered as subdomain, and 'example' as domain
        ("https://localhost.example", False),
        ("https://sub1.localhost.example", False),
        ("localhost", False),
        ("https://125.0.0.0", False),
        ("com", False),
        ("*", False),
        ("//", False),
        ("", False),
        ("https:///integration/ckeditor", False),
    ],
    scope="function",
)
def test_url_is_validated_as_expected(input_url, expected_result):
    output_result = URLHandler(url=input_url).url_is_usable()
    assert output_result == expected_result


@pytest.mark.parametrize(
    argnames=("input_parsed_url", "expected_path"),
    argvalues=(
        (
            ParseResult(
                scheme="https",
                netloc="sub1.domain.com",
                path="/demo",
                params="",
                query="",
                fragment="",
            ),
            "/demo",
        ),
        (
            ParseResult(
                scheme="",
                netloc="domain.com",
                path="/demo",
                params="",
                query="courses=1",
                fragment="",
            ),
            "/demo",
        ),
        (
            ParseResult(
                scheme="",
                netloc="domain.com",
                path="/demo/",
                params="",
                query="",
                fragment="",
            ),
            "/demo/",
        ),
        (
            ParseResult(
                scheme="https",
                netloc="",
                path="/demo",
                params="",
                query="courses=1",
                fragment="",
            ),
            "/demo",
        ),
        (
            ParseResult(
                scheme="https",
                netloc="125.0.0.0",
                path="/demo",
                params="",
                query="",
                fragment="",
            ),
            "/demo",
        ),
        (
            ParseResult(
                scheme="https",
                netloc="domain.com",
                path="",
                params="",
                query="courses=1",
                fragment="",
            ),
            None,
        ),
        (
            ParseResult(scheme="", netloc="", path="domain", params="", query="", fragment=""),
            None,
        ),
    ),
    scope="function",
)
def test_path_is_processed_as_expected(input_parsed_url, expected_path):
    output_path = URLHandler.process_path(parsed_url=input_parsed_url)
    assert output_path == expected_path


@pytest.mark.parametrize(
    argnames=("input_url", "num_subdomains", "expected_subdomains_list"),
    argvalues=(
        ("https://sub2.sub1.example.com/path?query#fragment", 1, ["sub1"]),
        ("https://sub2.sub1.example.com/path?query#fragment", 2, ["sub2", "sub1"]),
        ("https://sub2.sub1.example.com/path?query#fragment", 3, ["sub2", "sub1"]),
        ("www.example.com", 1, ["www"]),
        ("www.example.com", 0, []),
        # Note that com.uk isn't a valid TLD (but it is a usable URL), so it is expected that the
        # funcition would consider 'example' as subdomain, 'com' as domain, and 'uk' as suffix
        ("https://example.com.uk", 1, ["example"]),
        ("https://example.com", 1, []),
        ("https://user@example.co.uk:8000", 1, []),
        ("125.0.0.0", 2, []),
        # Note that 125.0.0.0.0 isn't a valid IP, so the last 0 will be treated as base domain, and
        # the remaining part as subdomain
        ("125.0.0.0.0", 4, ["125", "0", "0", "0"]),
        ("", 1, []),
    ),
    scope="function",
)
def test_n_subdomains_are_gotten_correctly(input_url, num_subdomains, expected_subdomains_list):
    output_subdomains_list = URLHandler(url=input_url).get_n_subdomains(
        num_subdomains=num_subdomains
    )
    assert output_subdomains_list == expected_subdomains_list


@pytest.mark.parametrize(
    argnames=("input_url", "num_subdomains"),
    argvalues=(("https://sub2.sub1.example.com/path?query#fragment", -1),),
    scope="function",
)
def test_exception_is_raised_for_num_subdomain_smaller_than_0(input_url, num_subdomains):
    with pytest.raises(ValueError):
        URLHandler(url=input_url).get_n_subdomains(num_subdomains=num_subdomains)


@pytest.mark.parametrize(
    argnames=("input_domains", "input_paths", "expected_joined_domain_and_path_list"),
    argvalues=(
        (["example.com"], ["here"], ["example.com", "example.com/here"]),
        (
            ["example.com", "this.example.com"],
            ["here"],
            ["example.com", "example.com/here", "this.example.com", "this.example.com/here"],
        ),
        (
            ["example.com", "this.example.com"],
            ["here", "here/there"],
            [
                "example.com",
                "example.com/here",
                "example.com/here/there",
                "this.example.com",
                "this.example.com/here",
                "this.example.com/here/there",
            ],
        ),
        (["example.com", "this.example.com"], [], ["example.com", "this.example.com"]),
    ),
    scope="function",
)
def test_domain_and_path_are_joined_properly(
    input_domains, input_paths, expected_joined_domain_and_path_list
):
    output_joined_domain_and_path_list = URLHandler.join_domain_with_path(
        domains=input_domains, paths=input_paths
    )

    assert sorted(output_joined_domain_and_path_list) == sorted(
        expected_joined_domain_and_path_list
    )


@pytest.mark.parametrize(
    argnames=("input_base_domain", "input_suffix", "expected_base_domain_with_suffix"),
    argvalues=(
        ("domain", "es", "domain.es"),
        ("domain", "co.uk", "domain.co.uk"),
        ("*", "es", "*.es"),
        ("domain", None, "domain"),
        ("125.0.0.0", None, "125.0.0.0"),
        ("*", None, "*"),
        (None, "es", "es"),
        (None, None, None),
    ),
    scope="function",
)
def test_base_domain_is_joined_with_suffix_properly(
    input_base_domain, input_suffix, expected_base_domain_with_suffix
):
    output_base_domain_with_suffix = URLHandler.join_base_domain_with_suffix(
        base_domain=input_base_domain, suffix=input_suffix
    )
    assert output_base_domain_with_suffix == expected_base_domain_with_suffix


@pytest.mark.parametrize(
    argnames=("input_subdomains", "input_base_domain_with_suffix", "expected_result"),
    argvalues=(
        (["sub2", "sub1"], "domain.es", "sub2.sub1.domain.es"),
        (["sub1"], "domain.es", "sub1.domain.es"),
        ([], "domain.es", "domain.es"),
        (["sub2", "sub1"], "*", "sub2.sub1.*"),
        ([], "domain", "domain"),
        ([], "*", "*"),
        ([], "es", "es"),
        (["sub2", "sub1"], None, "sub2.sub1"),
        ([], None, None),
    ),
    scope="function",
)
def test_subdomains_are_joined_with_base_domain_and_suffix_properly(
    input_subdomains, input_base_domain_with_suffix, expected_result
):
    output_result = URLHandler.join_subdomains_with_base_domain_and_suffix(
        subdomains=input_subdomains, base_domain_with_suffix=input_base_domain_with_suffix
    )
    assert expected_result == output_result


@pytest.mark.parametrize(
    argnames=("input_url", "num_subdomains", "expected_partial_domain"),
    argvalues=[
        ("https://sub2.sub1.example.com/path?query#fragment", 0, "example.com"),
        ("https://sub2.sub1.example.com", 1, "sub1.example.com"),
        ("https://sub2.sub1.example.com", 2, "sub2.sub1.example.com"),
        ("https://sub2.sub1.example.com", 3, "sub2.sub1.example.com"),
        ("https://sub2.sub1.example.com", None, "sub2.sub1.example.com"),
        ("https://example.com", 1, "example.com"),
        ("https://example.com", None, "example.com"),
        ("https://example.co.uk", 0, "example.co.uk"),
        ("https://user@example.co.uk:8000", 0, "example.co.uk"),
        # Note that com.uk isn't a valid TLD (but it is a usable URL), so it is expected that the
        # function would consider 'example' as subdomain, 'com' as domain, and 'uk' as suffix
        ("https://example.com.uk", 0, "com.uk"),
        ("www.example.com", 0, "example.com"),
        # Notice that this URL is not usable because it doesn't contain suffix, so 'localhost' will be
        # be considered as subdomain, and 'example' as domain
        ("https://localhost.example", 0, "example"),
        ("https://sub1.localhost.example", 1, "localhost.example"),
        ("https://sub1.localhost.example", 2, "sub1.localhost.example"),
        ("localhost", 0, "localhost"),
        ("localhost", None, "localhost"),
        ("localhost.example", None, "localhost.example"),
        ("https://125.0.0.0", 1, "125.0.0.0"),
        ("com", 0, "com"),
        ("*", 0, "*"),
        ("//", 0, None),
        ("", 1, None),
        ("https:///integration/ckeditor", 0, None),
        ("https:///integration/ckeditor", 1, None),
        ("https:///integration/ckeditor", None, None),
    ],
    scope="function",
)
def test_partial_domain_is_gotten_correctly(
    input_url,
    num_subdomains,
    expected_partial_domain,
):
    output_partial_domain = URLHandler(url=input_url).get_partial_domain(
        num_subdomains=num_subdomains
    )

    assert output_partial_domain == expected_partial_domain


@pytest.mark.parametrize(
    argnames=("input_url", "expected_result"),
    argvalues=[
        ("https://sub1.example.com/path?query#fragment", False),
        ("https://125.0.0.0", True),
        ("https://277.0.0.0", False),
        ("125.0.0.0:8000", True),
        ("user@125.0.0.0:8000", True),
        ("125.0.0.0.0", False),
        ("1", False),
        ("0", False),
        ("", False),
        ("https:///integration/ckeditor", False),
    ],
    scope="function",
)
def test_contains_ip_evaluates_valid_and_invalid_url_correctly(input_url, expected_result):
    test_containing_ip_evaluation = URLHandler(url=input_url).contains_ip()
    assert test_containing_ip_evaluation == expected_result


@pytest.mark.parametrize(
    argnames=("input_url", "expected_exploded_domains"),
    argvalues=[
        (
            "http://sub2.sub1.example.com?query#fragment",
            ["example.com", "sub1.example.com", "sub2.sub1.example.com"],
        ),
        ("https://sub1.example.com/path?query#fragment", ["example.com", "sub1.example.com"]),
        ("https://example.com", ["example.com"]),
        ("https://example.co.uk", ["example.co.uk"]),
        ("https://user@example.co.uk:8000", ["example.co.uk"]),
        ("https://example.com.uk", ["com.uk", "example.com.uk"]),
        ("www.example.com", ["example.com", "www.example.com"]),
        ("https://localhost.example", ["example", "localhost.example"]),
        ("localhost", ["localhost"]),
        ("125.0.0.0", ["125.0.0.0"]),
        ("com", ["com"]),
        ("*", ["*"]),
        ("//", []),
        ("", []),
        ("https:///js", []),
        ("https:///integration/ckeditor", []),
    ],
    scope="function",
)
def test_domains_are_exploded_properly(input_url, expected_exploded_domains):
    output_exploded_domains = URLHandler(url=input_url).explode_domain()
    assert output_exploded_domains == expected_exploded_domains


@pytest.mark.parametrize(
    argnames=("input_data", "input_depth", "expected_exploded_paths"),
    argvalues=(
        ("/moodle", 1, ["moodle"]),
        ("/moodle", 2, ["moodle"]),
        ("/moodle/", 1, ["moodle"]),
        ("/moodle/this", 1, ["moodle"]),
        ("/moodle/this/", 1, ["moodle"]),
        ("/moodle/this", 2, ["moodle", "moodle/this"]),
        ("/moodle/this/", 2, ["moodle", "moodle/this"]),
        ("/moodle/this", 3, ["moodle", "moodle/this"]),
        ("moodle/this", 1, ["moodle"]),
        ("/moodle/subpath1/subpath2", 2, ["moodle", "moodle/subpath1"]),
        (
            "/moodle/subpath1/subpath2",
            3,
            ["moodle", "moodle/subpath1", "moodle/subpath1/subpath2"],
        ),
        ("/", 1, []),
        ("/", 2, []),
        ("", 1, []),
        ("", 2, []),
    ),
    scope="function",
)
def test_explode_path(input_data, input_depth, expected_exploded_paths):
    output_exploded_paths = URLHandler.explode_path(path=input_data, max_depth=input_depth)

    assert output_exploded_paths == expected_exploded_paths



import json
import os
from typing import Dict, List

import pandas as pd
import pytest
import requests
from moto import mock_secretsmanager
from santoku.aws.secretsmanager import SecretsManagerHandler
from santoku.exceptions import MissingEnvironmentVariables
from santoku.salesforce.lightning import (
    LightningRestApiHandler,
    RequestMethodError,
    SalesforceObjectError,
    SalesforceObjectFieldError,
)

"""
Note: This class necessitates a Salesforce instance up and running in order to pass the tests.
We at Wiris execute this tests against our own private Salesforce testing sandbox instance.
In order to pass those tests, simply setup Salesforce and pass the credentials below.
Be warned that the fixtures in the tests can and will create and destroy Salesforce objects in
your instance. We do try to only create sobjects with narrow names and only destroy the ones we
create, so that we don't affect the rest of the Salesforce environment. A good practice, however,
would be to periodically clean that instance in order to reduce chance of existing data leaking
into the tests.
"""


credentials_keys = [
    "DATA_SCIENCE_SALESFORCE_SANDBOX_AUTH_URL",
    "DATA_SCIENCE_SALESFORCE_SANDBOX_USR",
    "DATA_SCIENCE_SALESFORCE_SANDBOX_PSW",
    "DATA_SCIENCE_SALESFORCE_SANDBOX_CLIENT_USR",
    "DATA_SCIENCE_SALESFORCE_SANDBOX_CLIENT_PSW",
]

if not all(key in os.environ for key in credentials_keys):
    raise MissingEnvironmentVariables("Salesforce credentials environment variables are missing.")


@pytest.fixture(scope="class")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "eu-west-1"


@pytest.fixture(scope="class")
def secrets_manager(aws_credentials):
    with mock_secretsmanager():
        secrets_manager = SecretsManagerHandler()
        yield secrets_manager


@pytest.fixture(scope="class")
def sf_credentials():
    return {
        "AUTH_URL": os.environ["DATA_SCIENCE_SALESFORCE_SANDBOX_AUTH_URL"],
        "USR": os.environ["DATA_SCIENCE_SALESFORCE_SANDBOX_USR"],
        "PSW": os.environ["DATA_SCIENCE_SALESFORCE_SANDBOX_PSW"],
        "CLIENT_USR": os.environ["DATA_SCIENCE_SALESFORCE_SANDBOX_CLIENT_USR"],
        "CLIENT_PSW": os.environ["DATA_SCIENCE_SALESFORCE_SANDBOX_CLIENT_PSW"],
    }


@pytest.fixture(scope="function")
def sf_credentials_secret(secrets_manager, sf_credentials, request):
    secret_name = "test/sf_credentials_secret"
    secrets_manager.client.create_secret(Name=secret_name, SecretString=json.dumps(sf_credentials))

    yield secret_name

    def teardown() -> None:
        secrets_manager.client.delete_secret(SecretId=secret_name, ForceDeleteWithoutRecovery=True)

    request.addfinalizer(teardown)


@pytest.fixture(scope="class")
def api_handler(sf_credentials):
    return LightningRestApiHandler(
        auth_url=sf_credentials["AUTH_URL"],
        username=sf_credentials["USR"],
        password=sf_credentials["PSW"],
        client_id=sf_credentials["CLIENT_USR"],
        client_secret=sf_credentials["CLIENT_PSW"],
    )


@pytest.fixture(scope="function")
def contact_payloads():
    # Using Alice & Bob notation. More info at https://en.wikipedia.org/wiki/Alice_and_Bob
    names = ["Alice", "Bob", "Carol", "David"]
    last_names = ["Ackerman", "Bayes", "Cauchy", "Dijkstra"]
    emails = ["alice@test.com", "bob@test.com", "carol@test.com", "david@test.com"]

    payloads = []
    for contact in zip(names, last_names, emails):
        payloads.append({"FirstName": contact[0], "LastName": contact[1], "Email": contact[2]})

    return payloads


@pytest.fixture(scope="function")
def delete_record(api_handler):
    def _delete_record(sobject: str, record_id: str) -> None:
        api_handler.do_request(
            method="DELETE",
            path=f"sobjects/{sobject}/{record_id}",
        )

    return _delete_record


@pytest.fixture(scope="function")
def insert_record(api_handler):
    def _insert_record(sobject: str, payload: List[Dict[str, str]]) -> None:
        response_text = api_handler.do_request(
            method="POST", path=f"sobjects/{sobject}", payload=payload
        )
        record_id = json.loads(response_text)["id"]
        return record_id

    return _insert_record


@pytest.fixture(scope="function")
def contacts(contact_payloads, insert_record, delete_record, request):
    created_record_ids = []
    for payload in contact_payloads:
        created_record_ids.append(insert_record(sobject="Contact", payload=payload))
    yield created_record_ids

    def teardown() -> None:
        for record_id in created_record_ids:
            try:
                delete_record(sobject="Contact", record_id=record_id)
            except:
                pass

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def response():
    return [
        {
            "Quantity__c": 1.0,
            "Product__r": {
                "attributes": {
                    "type": "Product2",
                    "url": "/services/data/v47.0/sobjects/Product2/01t0N000009sBmaQAE",
                },
                "Recurrency__c": "Subscription",
            },
            "Quote": {
                "attributes": {
                    "type": "Quote",
                    "url": "/services/data/v47.0/sobjects/Quote/0Q00N000002AO0NSAW",
                },
                "Account": {
                    "attributes": {
                        "type": "Account",
                        "url": "/services/data/v47.0/sobjects/Account/0010N00004j7w2KQAQ",
                    },
                    "Name": "Viewpoint School - Upper & Middle School",
                },
            },
        },
        {
            "Quantity__c": 2.0,
            "Product__r": {
                "attributes": {
                    "type": "Product2",
                    "url": "/services/data/v47.0/sobjects/Product2/01t0N000009sBmaQAE",
                },
                "Recurrency__c": "Renewal",
            },
            "Quote": {
                "attributes": {
                    "type": "Quote",
                    "url": "/services/data/v47.0/sobjects/Quote/0Q00N000002AO0NSAW",
                },
                "Account": {
                    "attributes": {
                        "type": "Account",
                        "url": "/services/data/v47.0/sobjects/Account/0010N00004j7w2KQAQ",
                    },
                    "Name": "Viewpoint School",
                },
            },
        },
    ]


@pytest.fixture(scope="function")
def reference():

    return pd.DataFrame(
        {
            "quantity": [1.0, 2.0],
            "product_recurrency": ["Subscription", "Renewal"],
            "quote_account_name": ["Viewpoint School - Upper & Middle School", "Viewpoint School"],
        }
    )


@pytest.fixture(scope="function")
def contacts_df(contact_payloads):

    df = pd.DataFrame(contact_payloads)

    # Sort dataframe to ensure determinism
    return df.sort_values(by=list(df.columns), ignore_index=True)


class TestLightningRestApiHandler:
    def test_wrong_credentials(self, sf_credentials, contact_payloads):
        # Connect Salesforce with wrong credentials. Failure expected.
        api_handler = LightningRestApiHandler(
            auth_url=sf_credentials["AUTH_URL"],
            username="false_username",
            password="false_password",
            client_id=sf_credentials["CLIENT_USR"],
            client_secret=sf_credentials["CLIENT_PSW"],
        )
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=contact_payloads[0],
            )

        api_handler = LightningRestApiHandler(
            auth_url=sf_credentials["AUTH_URL"],
            username=sf_credentials["USR"],
            password=sf_credentials["PSW"],
            client_id="false_client_id",
            client_secret="false_client_secret",
        )
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=contact_payloads[0],
            )

    def test_salesforce_object_required_fields(self, api_handler):
        # Test inserting an invalid field. Failure expected.
        invalid_field = "InvalidField"
        bad_contact_payload = {invalid_field: "InvalidValue"}
        expected_message = f"`{invalid_field}` isn't a valid field."
        with pytest.raises(SalesforceObjectFieldError, match=expected_message):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=bad_contact_payload,
            )

        # Test inserting a contact without a required field. Failure expected.
        bad_contact_payload = {
            "FirstName": "Chuck",
            "Email": "chuck@test.com",
        }

        missing_field = "LastName"
        expected_message = (
            f"`{missing_field}` is a required field and does not appear in the payload."
        )
        with pytest.raises(SalesforceObjectFieldError, match=expected_message):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=bad_contact_payload,
            )

        # Test inserting a contact with an empty required field. Failure expected.
        bad_contact_payload[missing_field] = ""
        expected_message = f"`{missing_field}` is a required field and must not be empty."
        with pytest.raises(SalesforceObjectFieldError, match=expected_message):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=bad_contact_payload,
            )

    def test_contact_insertion(self, api_handler, contact_payloads, delete_record):
        created_record_ids = []

        # Insert n Contacts that do not exist. Success expected.
        for contact_payload in contact_payloads:
            response_text = api_handler.do_request(
                method="POST", path="sobjects/Contact", payload=contact_payload
            )
            response = json.loads(response_text)
            created_record_ids.append(response["id"])
            assert response["success"]

        # Insert a Contact that already exist. Failure expected.
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.do_request(
                method="POST",
                path="sobjects/Contact",
                payload=contact_payloads[0],
            )

        # Insert a Contact that already exist with a new email. Success expected.
        new_contact_payload = contact_payloads[0].copy()
        new_contact_payload["Email"] = "new.email@example.com"
        response_text = api_handler.do_request(
            method="POST",
            path="sobjects/Contact",
            payload=new_contact_payload,
        )
        response = json.loads(response_text)
        created_record_ids.append(response["id"])
        assert response["success"]

        # Remove created records.
        for record_id in created_record_ids:
            delete_record(sobject="Contact", record_id=record_id)

    def test_init_handler_from_secrets_manager(
        self, sf_credentials_secret, contact_payloads, delete_record
    ):
        # Initialize the handler from secrets using the default secret keys by convention and insert
        # a Contact that does not exist. Success expected.
        api_handler = LightningRestApiHandler.from_aws_secrets_manager(
            secret_name=sf_credentials_secret
        )
        response_text = api_handler.do_request(
            method="POST",
            path="sobjects/Contact",
            payload=contact_payloads[0],
        )
        response = json.loads(response_text)
        assert response["success"]

        # Remove created records.
        delete_record(sobject="Contact", record_id=response["id"])

    def test_different_query_syntaxes(self, api_handler):
        # Do a query written in uppercase. Success expected.
        obtained_contacts = api_handler.do_query_with_SOQL("SELECT ID, NAME FROM CONTACT")
        assert obtained_contacts.empty

        # Do a query written in lowercase. Success expected.
        obtained_contacts = api_handler.do_query_with_SOQL("select id, name from contact")
        assert obtained_contacts.empty

        # Do a query to a non-existent object. Failure expected.
        wrong_object_name = "Contacts"
        expected_message = f"{wrong_object_name} isn't a valid object"
        with pytest.raises(SalesforceObjectError, match=expected_message):
            obtained_contacts = api_handler.do_query_with_SOQL(
                f"SELECT Id, Name From {wrong_object_name}"
            )

    def test_contact_query(self, api_handler, contact_payloads, contacts, contacts_df):

        # Read the Contacts inserted with SOQL. Success expected.
        obtained_contacts = api_handler.do_query_with_SOQL(
            "SELECT FirstName, LastName, Email FROM Contact"
        )

        # Sort dataframe to ensure determinism
        obtained_contacts.sort_values(
            by=list(obtained_contacts.columns), inplace=True, ignore_index=True
        )

        assert obtained_contacts.equals(other=contacts_df)

        # Read a specific contact with SOQL by FirstName. Success expected.
        first_name = contact_payloads[0]["FirstName"]

        obtained_contacts = api_handler.do_query_with_SOQL(
            f"SELECT FirstName, LastName, Email FROM contact WHERE FirstName = '{first_name}'"
        )
        obtained_contacts.sort_values(
            by=list(obtained_contacts.columns), inplace=True, ignore_index=True
        )

        expected_contacts = contacts_df[contacts_df["FirstName"] == first_name].sort_values(
            by=list(contacts_df.columns), ignore_index=True
        )

        assert obtained_contacts.equals(other=expected_contacts)

        # Query a contact that does not exists with SOQL. Success expected.
        obtained_contacts = api_handler.do_query_with_SOQL(
            f"SELECT FirstName, LastName, Email FROM contact WHERE FirstName = 'dummyname'"
        )
        assert obtained_contacts.empty

    def test_contact_modification(self, api_handler, contact_payloads, contacts, contacts_df):
        # Modify an existing contact's Name. Success expected.
        new_first_name = "Ken"
        new_last_name = "Williams"
        new_contact_payload = {"FirstName": new_first_name, "LastName": new_last_name}
        contact_id = contacts[0]
        api_handler.do_request(
            method="PATCH",
            path=f"sobjects/Contact/{contact_id}",
            payload=new_contact_payload,
        )

        contacts_df.replace(
            {"FirstName": {contact_payloads[0]["FirstName"]: new_first_name}}, inplace=True
        )
        contacts_df.replace(
            {"LastName": {contact_payloads[0]["LastName"]: new_last_name}}, inplace=True
        )

        obtained_contact = api_handler.do_query_with_SOQL(
            f"SELECT FirstName, LastName, Email from Contact WHERE FirstName = '{new_first_name}'"
        )

        obtained_contact.sort_values(
            by=list(obtained_contact.columns), inplace=True, ignore_index=True
        )

        contacts_df = contacts_df[contacts_df["FirstName"] == new_first_name].sort_values(
            by=list(contacts_df.columns), ignore_index=True
        )

        assert obtained_contact.equals(other=contacts_df)

        # Modify a Contact that does not exist. Failure expected.
        new_contact_payload = {"FirstName": "NEWNAME"}
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.do_request(
                method="PATCH",
                path="sobjects/Contact/WRONGID",
                payload=new_contact_payload,
            )

    def test_contact_deletion(self, api_handler, contacts):
        # Delete an existing Contact. Success expected.
        api_handler.do_request(
            method="DELETE",
            path=f"sobjects/Contact/{contacts[0]}",
        )

        obtained_contacts = api_handler.do_query_with_SOQL(
            f"SELECT Name from Contact WHERE Id = '{contacts[0]}'"
        )
        assert obtained_contacts.empty

        # Delete a Contact that does not exist. Failure expected.
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.do_request(
                method="DELETE",
                path=f"sobjects/Contact/{contacts[0]}",
            )

    def test_contact_insertion_high_level(self, api_handler, contact_payloads, delete_record):
        created_record_ids = []

        # Insert n Contacts that do not exist. Success expected.
        for contact_payload in contact_payloads:
            response_text = api_handler.insert_record(sobject="Contact", payload=contact_payload)
            response = json.loads(response_text)
            created_record_ids.append(response["id"])
            assert response["success"]

        # Insert a Contact that already exist. Failure expected.
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.insert_record(
                sobject="Contact",
                payload=contact_payloads[0],
            )

        # Insert a Contact that already exist with a new email. Success expected.
        new_contact_payload = contact_payloads[0].copy()
        new_contact_payload["Email"] = "new.email@example.com"
        response_text = api_handler.insert_record(sobject="Contact", payload=new_contact_payload)
        response = json.loads(response_text)
        created_record_ids.append(response["id"])
        assert response["success"]

        # Insert a Contact that already exist. Failure expected.
        with pytest.raises(requests.exceptions.RequestException):
            response = api_handler.insert_record(sobject="Contact", payload=contact_payloads[0])
            assert response

        # Remove created records.
        for record_id in created_record_ids:
            delete_record(sobject="Contact", record_id=record_id)

    def test_contact_modification_high_level(
        self, api_handler, contacts, contact_payloads, contacts_df
    ):
        # Modify an existing contact's Name. Success expected.
        new_first_name = "Ralph"
        new_last_name = "Alexander"
        new_contact_payload = {"FirstName": new_first_name, "LastName": new_last_name}
        contact_id = contacts[0]
        api_handler.modify_record(
            sobject="Contact",
            record_id=contact_id,
            payload=new_contact_payload,
        )

        obtained_contact = api_handler.do_query_with_SOQL(
            f"SELECT FirstName, LastName, Email FROM Contact WHERE FirstName = '{new_first_name}'"
        )

        obtained_contact.sort_values(by=list(contacts_df.columns), inplace=True, ignore_index=True)

        contacts_df.replace(
            {"FirstName": {contact_payloads[0]["FirstName"]: new_first_name}}, inplace=True
        )
        contacts_df.replace(
            {"LastName": {contact_payloads[0]["LastName"]: new_last_name}}, inplace=True
        )

        contacts_df = contacts_df[contacts_df["FirstName"] == new_first_name].sort_values(
            by=list(contacts_df.columns), ignore_index=True
        )

        # expected_contact_name = f"{new_first_name} {new_last_name}"
        assert obtained_contact.equals(contacts_df)

        # Modify a Contact that does not exist. Failure expected.
        new_contact_payload = {"FirstName": "NEWNAME"}
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.modify_record(
                sobject="Contact",
                record_id="WRONGID",
                payload=new_contact_payload,
            )

    def test_contact_deletion_high_level(self, api_handler, contacts):
        # Delete an existing Contact. Success expected.
        api_handler.delete_record(sobject="Contact", record_id=contacts[0])

        obtained_contacts = api_handler.do_query_with_SOQL(
            f"SELECT Name from Contact WHERE Id = '{contacts[0]}'"
        )
        assert obtained_contacts.empty

        obtained_contacts = api_handler.do_query_with_SOQL(
            f"SELECT Name from contact WHERE Id = '{contacts[1]}'"
        )
        assert len(obtained_contacts) == 1

        # Delete a Contact that does not exist. Failure expected.
        with pytest.raises(requests.exceptions.RequestException):
            api_handler.delete_record(sobject="Contact", record_id=contacts[0])

    def test_soql_response_to_dataframe(self, api_handler, response, reference):
        test_result = api_handler._soql_response_to_dataframe(
            response=response,
            drop_columns_containing="attributes",
            column_mapping={
                "Quantity__c": "quantity",
                "Product__r.Recurrency__c": "product_recurrency",
                "Quote.Account.Name": "quote_account_name",
            },
        )
        assert test_result.equals(reference)



import json
import os
from io import BytesIO, StringIO
from typing import Dict, List

import boto3
import pandas as pd
import pytest
from botocore import exceptions
from moto import mock_s3
from santoku.aws import utils
from santoku.aws.s3 import ManifestError, S3Handler

"""
TODO: this whole section might serve in the future as test suite for this library
      The idea is to use moto to locally simulate the aws services so that the suite can run without
      actually connecting to the aws cloud, thus saving $$ and reducing the latency of the tests.
"""


@pytest.fixture(scope="class")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


@pytest.fixture(scope="class")
def s3_handler(aws_credentials):
    with mock_s3():
        s3_handler = S3Handler()
        yield s3_handler


@pytest.fixture(scope="function")
def bucket(s3_handler, request):
    bucket_name = "test_bucket"
    try:
        s3_handler.resource.meta.client.head_bucket(Bucket=bucket_name)
    except exceptions.ClientError:
        pass
    else:
        raise EnvironmentError(f"{test_bucket} should not exist.")
    s3_handler.client.create_bucket(Bucket=bucket_name)

    yield bucket_name

    def teardown() -> None:
        bucket = s3_handler.resource.Bucket(bucket_name)
        bucket.delete()

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def delete_object_in_s3(bucket, s3_handler):
    def _delete_object_in_s3(key: str) -> None:
        s3_handler.client.delete_object(Bucket=bucket, Key=key)

    return _delete_object_in_s3


@pytest.fixture(scope="function")
def put_object_to_s3(bucket, s3_handler):
    def _put_object_to_s3(object_name: str, content: str) -> None:
        s3_handler.client.put_object(Body=content, Bucket=bucket, Key=object_name)

    yield _put_object_to_s3


@pytest.fixture(scope="function")
def files_with_no_common_prefix(put_object_to_s3, delete_object_in_s3, request):
    files = {
        "no_prefix_object.json": "no_prefix_object content",
    }

    for file_name, content in files.items():
        put_object_to_s3(object_name=file_name, content=content)
    yield files

    def teardown() -> None:
        try:
            for file_name in files:
                delete_object_in_s3(key=file_name)
        except:
            pass

    request.addfinalizer(teardown)


@pytest.fixture(scope="class")
def prefix():
    return "test_prefix"


@pytest.fixture(scope="function")
def files_with_common_prefix(put_object_to_s3, prefix, delete_object_in_s3, request):
    files = {
        f"{prefix}_object_1.json": f"{prefix}_object_1 content",
        f"{prefix}_object_2.json": f"{prefix}_object_2 content",
    }

    for file_name, content in files.items():
        put_object_to_s3(object_name=file_name, content=content)
    yield files

    def teardown() -> None:
        for file_name in files:
            try:
                delete_object_in_s3(key=file_name)
            except:
                pass

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def fixture_objects(files_with_no_common_prefix, files_with_common_prefix):
    fixture_objects = files_with_no_common_prefix.copy()
    fixture_objects.update(files_with_common_prefix)
    return fixture_objects


@pytest.fixture(scope="function")
def dataframe():
    data = {
        "Column1": ["Value11", "Value21"],
        "Column2": ["Value12", "Value22"],
    }
    return pd.DataFrame(data)


class TestS3Handler:
    def test_get_uri(self, bucket, s3_handler):
        file_name = "test_file.test"

        # Bucket without folder nor file. Success expected.
        expected_path = f"s3://{bucket}/"
        obtained_path = s3_handler.get_uri(bucket=bucket)
        assert obtained_path == expected_path

        # File in a bucket without folder.
        expected_path = f"s3://{bucket}/{file_name}"
        obtained_path = s3_handler.get_uri(bucket=bucket, file_name=file_name)
        assert obtained_path == expected_path

        # Folder in a bucket without file. Success expected.
        folder = "folder"
        expected_path = f"s3://{bucket}/{folder}/"
        obtained_path = s3_handler.get_uri(
            bucket=bucket,
            folder_path=folder,
        )
        assert obtained_path == expected_path

        # Folder in a bucket with file. Success expected.
        expected_path = f"s3://{bucket}/{folder}/{file_name}"
        obtained_path = s3_handler.get_uri(
            bucket=bucket,
            folder_path=folder,
            file_name=file_name,
        )
        assert obtained_path == expected_path

        # Folder in a bucket with a / at the begining. Success expected.
        folder = "/folder"
        expected_path = f"s3://{bucket}{folder}/{file_name}"
        obtained_path = s3_handler.get_uri(
            bucket=bucket,
            folder_path=folder,
            file_name=file_name,
        )
        assert obtained_path == expected_path

        # Folder in a bucket with subfolder. Success expected.
        folder = "folder/subfolder"
        expected_path = f"s3://{bucket}/{folder}/{file_name}"
        obtained_path = s3_handler.get_uri(
            bucket=bucket,
            folder_path=folder,
            file_name=file_name,
        )
        assert obtained_path == expected_path

    def test_list_objects(
        self, bucket, s3_handler, prefix, files_with_no_common_prefix, files_with_common_prefix
    ):
        # List objects by prefix. Success expected.
        args = {
            "Prefix": prefix,
        }
        common_prefix_object_keys = [object_key for object_key in files_with_common_prefix]
        obtained_objects = list(s3_handler.list_objects(bucket=bucket, **args))
        assert obtained_objects == common_prefix_object_keys

        # List objects starting from a specific key.
        args = {
            "StartAfter": common_prefix_object_keys[0],
        }
        obtained_objects = list(s3_handler.list_objects(bucket=bucket, **args))
        assert obtained_objects == common_prefix_object_keys[1:]

        # List all objects. Success expected.
        expected_objects = set(list(files_with_no_common_prefix.keys()) + common_prefix_object_keys)
        obtained_objects = set(s3_handler.list_objects(bucket=bucket))
        assert obtained_objects == expected_objects

    def test_check_object_exists(self, bucket, s3_handler, fixture_objects):
        # Test that an object exist in the bucket. Success expected.
        obtained_results = [
            s3_handler.check_object_exists(object_key=object_key, bucket=bucket)
            for object_key in fixture_objects
        ]
        assert all(obtained_results)

        # Test that an object does not exist in the bucket. Success expected.
        obtained_result = s3_handler.check_object_exists(
            object_key="false_object.json", bucket=bucket
        )
        assert not obtained_result

    def test_read_object_content(self, bucket, s3_handler, fixture_objects):
        # Test reading existing objects. Sucess expected.
        for object_key, content in fixture_objects.items():
            obtained_content = s3_handler.read_object_content(bucket=bucket, object_key=object_key)
            expected_content = content
            assert obtained_content == expected_content

        # Test reading an object that does not exist. Failure expected.
        with pytest.raises(exceptions.ClientError) as e:
            obtained_content = s3_handler.read_object_content(
                bucket=bucket, object_key="false_object.json"
            )

    def test_put_object(self, bucket, s3_handler, files_with_no_common_prefix, delete_object_in_s3):
        # Put content to a new object. Success expected.
        new_object_key = "new_object.json"
        content_to_write = "new_object content"
        s3_handler.put_object(bucket=bucket, object_key=new_object_key, content=content_to_write)

        read_object = s3_handler.resource.Object(bucket_name=bucket, key=new_object_key)
        assert read_object

        content_read = read_object.get()["Body"].read().decode("utf-8")
        assert content_read == content_to_write

        # Put content to an object that already exist. Verify the content is overwriten.
        # Success expected.
        object_key = list(files_with_no_common_prefix.keys())[0]
        original_content = files_with_no_common_prefix[object_key]
        read_object = s3_handler.resource.Object(bucket_name=bucket, key=object_key)
        content_read = read_object.get()["Body"].read().decode("utf-8")
        assert content_read == original_content

        content_to_write = "New content"
        s3_handler.put_object(bucket=bucket, object_key=object_key, content=content_to_write)

        read_object = s3_handler.resource.Object(bucket_name=bucket, key=object_key)
        content_read = read_object.get()["Body"].read().decode("utf-8")
        assert content_read == content_to_write

        # Remove the new object.
        delete_object_in_s3(key=new_object_key)

    def test_delete_object(self, bucket, s3_handler, files_with_no_common_prefix):
        # Validate that an object does not exist anymore after being deleted. Failure expected.
        object_key = list(files_with_no_common_prefix.keys())[0]
        s3_handler.resource.Object(bucket_name=bucket, key=object_key).get()
        s3_handler.delete_object(bucket=bucket, object_key=object_key)

        with pytest.raises(exceptions.ClientError) as e:
            s3_handler.resource.Object(bucket_name=bucket, key=object_key).get()

    def test_write_dataframe_to_csv_object(
        self, bucket, s3_handler, delete_object_in_s3, dataframe
    ):
        # Validate that a dataframe can be written and read correctly in csv format.
        # Success expected.
        new_object_key = "dataframe.csv"
        s3_handler.write_dataframe_to_csv_object(
            bucket=bucket, object_key=new_object_key, dataframe=dataframe
        )

        read_object = s3_handler.resource.Object(bucket_name=bucket, key=new_object_key)
        object_content = read_object.get()["Body"].read().decode("utf-8")
        obtained_df = pd.read_csv(filepath_or_buffer=StringIO(object_content))
        assert obtained_df.equals(dataframe)

        # Remove the new object.
        delete_object_in_s3(key=new_object_key)

    def test_write_dataframe_to_parquet_object(
        self, bucket, s3_handler, delete_object_in_s3, dataframe
    ):
        # Validate that a dataframe can be written and read correctly in parquet format.
        # Success expected.
        new_object_key = "dataframe.parquet"
        s3_handler.write_dataframe_to_parquet_object(
            bucket=bucket, object_key=new_object_key, dataframe=dataframe
        )

        read_object = s3_handler.resource.Object(bucket_name=bucket, key=new_object_key)
        object_content = read_object.get()["Body"].read()
        obtained_df = pd.read_parquet(path=BytesIO(object_content))
        assert obtained_df.equals(dataframe)

        # Remove the new object.
        delete_object_in_s3(key=new_object_key)

    def test_generate_quicksight_manifest(
        self,
        bucket,
        s3_handler,
        files_with_no_common_prefix,
        prefix,
        files_with_common_prefix,
        delete_object_in_s3,
    ):
        manifest_key = "manifest.json"
        key_paths = [f"s3://{bucket}/{list(files_with_no_common_prefix.keys())[0]}"]
        uri_prefix_paths = [f"s3://{bucket}/{prefix}"]
        upload_settings = {
            "format": "CSV",
            "delimiter": ",",
            "textqualifier": "'",
            "containsHeader": "True",
        }
        data = {
            "fileLocations": [{"URIs": key_paths}, {"URIPrefixes": uri_prefix_paths}],
            "globalUploadSettings": upload_settings,
        }

        # Test not specifying s3_uri_prefixes and s3_uris. Failure expected.
        with pytest.raises(ManifestError) as e:
            s3_handler.generate_quicksight_manifest(
                bucket=bucket,
                object_key=manifest_key,
                file_format=upload_settings["format"],
                delimiter=upload_settings["delimiter"],
                qualifier=upload_settings["textqualifier"],
                header_row=upload_settings["containsHeader"],
            )

        # Validate manifest file can be generated correctly. Success expected.
        expected_manifest = json.dumps(data, indent=4, sort_keys=True)
        s3_handler.generate_quicksight_manifest(
            bucket=bucket,
            object_key=manifest_key,
            s3_uris=key_paths,
            s3_uri_prefixes=uri_prefix_paths,
            file_format=upload_settings["format"],
            delimiter=upload_settings["delimiter"],
            qualifier=upload_settings["textqualifier"],
            header_row=upload_settings["containsHeader"],
        )

        obtained_manifest_bytes = (
            s3_handler.resource.Object(bucket_name=bucket, key=manifest_key).get()["Body"].read()
        )

        # Load the JSON to a Python list and dump it back out as formatted JSON
        obtained_manifest_dictionary = json.loads(obtained_manifest_bytes)
        obtained_manifest = json.dumps(obtained_manifest_dictionary, indent=4, sort_keys=True)
        assert obtained_manifest == expected_manifest

        # Remove the new object.
        delete_object_in_s3(key=manifest_key)



import json
import os
from io import BytesIO, StringIO
from typing import Any, Dict, Generator, List

import boto3
import pandas as pd
from botocore import exceptions
from santoku.aws import utils


class ManifestError(Exception):
    def __init__(self, message):
        super().__init__(message)


class S3Handler:
    """
    Class to manage input/output operations of Amazon S3 storage services.

    This class is intended to be run on AWS Glue jobs (Python Shell). S3 is organized in buckets,
    that contains a collection of files identified by a path. The connection to the elements of S3
    are done using the services classes Client and Resource of the library boto3. This class
    provides methods to interact with S3 elements and makes easy some usual operations.

    """

    def __init__(self, **kwargs):
        """ Instantiate the services classes. """
        self.client = boto3.client(service_name="s3", **kwargs)
        self.resource = boto3.resource(service_name="s3")

    @staticmethod
    def get_uri(bucket: str, folder_path: str = "", file_name: str = "") -> str:
        """
        Absolute S3 path (URI) of a file.

        Build an absolute S3 path containing the `bucket`, the `folder_path`
        and the `file_name` of the file. This method can be useful when some methods require an
        absolute path of the files.

        Parameters
        ----------
        bucket : str
            Name of bucket to build the absolute path with.
        folder_path : str, optional
            Relative folder path to build the absolute path with.
        file_name : str, optional
            Name of the file object to build the absolute path with.

        Returns
        -------
        str
            Absolut S3 URI of the file.

        """
        # Prevent os.path.join from considering folder_path as absolute path.
        if folder_path and folder_path[0] == "/":
            folder_path = folder_path[1:]
        return os.path.join("s3://", bucket, folder_path, file_name)

    def list_objects(self, bucket: str, **kwargs: Dict[str, str]) -> Generator[str, None, None]:
        """
        Get all objects in a specific location.

        Get the object keys located in the `bucket`. Yields an iterable
        with the found object keys in alphabetical order.

        Parameters
        ----------
        bucket : str
            Name of the bucket to iterate in.
        kwargs : Any
            Additional arguments for the used method boto3.client.list_objects_v2. Some usual
            arguments are Prefix (str), that filters those object keys that begin with the specified
            string, or StartAfter (str), that indicates where must S3 start listing from.

        Yields
        ------
        Generator[str, None, None]
            The object keys located in the `bucket` in alphabetical order.

        Notes
        -----
        This method will never add objects partially, or all objects will be added or none of them.
        If there are multiple write requests of the same object simultaneously, it overwrites all
        but the last object written. More information on list_objects_v2 method: [1].

        References
        ----------
        [1] :
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects_v2

        """
        args: Dict[str, Any] = {"Bucket": bucket}
        args.update(**kwargs)
        for result in utils.paginate(
            client=self.client, method=self.client.list_objects_v2.__name__, **args
        ):
            for contents in result["Contents"]:
                yield contents["Key"]

    def check_object_exists(self, object_key: str, bucket: str) -> bool:
        """
        Check whether an object exist.

        Return true if the object `object_key` exist in the `bucket`.

        Parameters
        ----------
        bucket : str
            Name of bucket to find the object.
        object_key : str
            Identifier of the object in S3 to find in the bucket.

        Returns
        ------
        bool
            True if `object_key` is inside the `bucket`.

        """
        try:
            self.resource.Object(bucket_name=bucket, key=object_key).load()
        except exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise
        return True

    def read_object_content(self, bucket: str, object_key: str, encoding="utf-8") -> str:
        """
        Get the content of a file.

        Return the content of the object `object_key` in the `bucket` with the specific encoding.
        If the object `object_key` does not exist, an exception will be raised.

        Parameters
        ----------
        bucket : str
            Name of the bucket to read from.
        object_key : str
            Identifier of the object in the bucket to read from.
        encoding : str, optional
            Type of encoding used in the content (the default is 'utf-8').

        Returns
        ------
        str
            The content of the object decoded.

        Raises
        ------
        botocore.exceptions.ClientError
            If the object called `object_key` does not exist in the `bucket`.

        """
        object_to_read = self.resource.Object(bucket_name=bucket, key=object_key)
        file_content = object_to_read.get()["Body"].read().decode(encoding)
        return file_content

    def put_object(self, bucket: str, object_key: str, content: bytes) -> None:
        """
        Write in a file in a specific location.

        Write `content` into an object `object_key` and upload it to the `bucket`. If the object
        already exist, its content will be overwriten.

        Parameters
        ----------
        bucket : str
            Name of the bucket to put the object in.
        object_key : str
            Identifier of the object to put into the bucket.
        content : bytes
            The contet to be writen into the object.

        Returns
        -------
        None

        """
        self.client.put_object(Body=content, Bucket=bucket, Key=object_key)
        return None

    def delete_object(self, bucket: str, object_key: str) -> None:
        """
        Remove a file.

        Delete an object `object_key` from the `bucket`. This method assumes the object `object_key`
        exists in the `bucket`.

        Parameters
        ----------
        bucket : str
            Name of the bucket from where the object will be deleted.
        object_key: str
            Identifier of the object to be delted.

        Returns
        -------
        None

        Raises
        ------
        botocore.exceptions.ClientError
            If the object called `object_key` does not exist in the `bucket`.

        """
        self.resource.Object(bucket_name=bucket, key=object_key).delete()
        return None

    def write_dataframe_to_csv_object(
        self,
        bucket: str,
        object_key: str,
        dataframe: pd.DataFrame,
        encoding: str = "utf-8",
        save_index: bool = False,
        **kwargs,
    ) -> None:
        """
        Put a dataframe into a csv file.

        Write the content of a pandas dataframe into a csv file and upload it to the `bucket`. If
        the object already exists, its content will be overwriten.

        Parameters
        ----------
        bucket : str
            Name of the bucket to upload the file.
        object_key : str
            Identifier of the file to be uploaded to the bucket.
        dataframe : pd.DataFrame
            Pandas dataframe that contains the content to be writen in the file.
        encoding : str, optional
            Type of encoding used in the content (the default is 'utf-8').
        save_index : bool, optional
            Whether to save the index as a column, (the default is 'False').

        Return
        ------
        None

        """
        # Get pandas dataframe as CSV bytes
        csv_buffer = StringIO()
        dataframe.to_csv(path_or_buf=csv_buffer, index=save_index, **kwargs)
        bytes_content = csv_buffer.getvalue().encode(encoding)
        self.put_object(bucket=bucket, object_key=object_key, content=bytes_content)

    def write_dataframe_to_parquet_object(
        self,
        bucket: str,
        object_key: str,
        dataframe: pd.DataFrame,
        compression: str = "snappy",
        **kwargs,
    ) -> None:
        """
        Put a dataframe into a parquet file.

        Write the content of a pandas dataframe into a parquet file and upload it to the `bucket`.
        If the object already exists, its content will be overwriten. The dataframe will be
        converted into parquet with `pyarrow` engine.

        Parameters
        ----------
        bucket : str
            Name of the bucket to upload the file.
        object_key : str
            Identifier of the file to be uploaded to the bucket.
        dataframe : pd.DataFrame
            Pandas dataframe that contains the content to be writen in the file.
        engine : bool, optional
            Parquet library to use, (the default is 'auto').

        Return
        ------
        None

        """
        parquet_buffer = BytesIO()
        engine = "pyarrow"
        dataframe.to_parquet(path=parquet_buffer, engine=engine, compression=compression, **kwargs)
        bytes_content = parquet_buffer.getvalue()
        self.put_object(bucket=bucket, object_key=object_key, content=bytes_content)

    def generate_quicksight_manifest(
        self,
        bucket: str,
        object_key: str,
        s3_uris: List[str] = None,
        s3_uri_prefixes: List[str] = None,
        file_format: str = None,
        delimiter: str = None,
        qualifier: str = None,
        header_row: bool = None,
    ):
        """
        Generates a QS manifest JSON file.

        Generates a QS manifest JSON file from a list of `S3 URIs` and/or `S3 URI prefixes` that is
        uploaded to `bucket`

        Parameters
        ----------
        bucket : str
            Name of the bucket to save the generated manifest.
        object_key : str
            Identifier with which the manifest will have in the bucket.
        s3_uris : List[str], optional
            List of S3 uris of the files that will be used.
        s3_uri_prefixes : List[str], optional
            List of S3 uri prefixes that will be used. The uri_prefixes filters those object keys
            that begin with the specified string. Notice that an uri_prefix is different than an S3
            prefix, an uri_prefix contains also the first part of an absolute path:
            's3://bucket_name/..'.
        format : str, optional
            Format of manifest files to be imported (the default is 'CSV').
        delimiter : str, optional
            File field delimiter (the default is ',').
        qualifier : str, optional
            Text qualifier used in the file to specify the beginning of a text.
        header_row : str, optional
            Boolean string that has value 'True' if the file has a header row (the default is
            'True').

        Return
        -----
        None

        Raises
        ------
        ManifestError
            If the file or prefix are not specified.

        Notes
        -----
        More information on the JSON format of the QuickSight manifest files: [1].

        References
        ----------
        [1] :
        https://docs.aws.amazon.com/quicksight/latest/user/supported-manifest-file-format.html

        """
        if s3_uri_prefixes is None and s3_uris is None:
            raise ManifestError("No file or prefix were specified.")

        # Build the JSON with the not None attributes.
        # Build the 'fileLocations' part.
        uri: Dict[str, Any] = {}
        if s3_uris is not None:
            uri["URIs"] = list(s3_uris)

        uri_prefixes: Dict[str, Any] = {}
        if s3_uri_prefixes is not None:
            uri_prefixes["URIPrefixes"] = list(s3_uri_prefixes)

        data: Dict[str, Any] = {"fileLocations": []}
        if s3_uris is not None:
            data["fileLocations"].append(uri)
        if s3_uri_prefixes is not None:
            data["fileLocations"].append(uri_prefixes)

        # Build the 'globalUploadSettings' part.
        upload_settings: Dict[str, Any] = {}
        if file_format is not None:
            upload_settings["format"] = file_format
        if delimiter is not None:
            upload_settings["delimiter"] = delimiter
        if qualifier is not None:
            upload_settings["textqualifier"] = qualifier
        if header_row is not None:
            upload_settings["containsHeader"] = header_row

        if upload_settings:
            data["globalUploadSettings"] = upload_settings
        json_content = json.dumps(data, indent=4, sort_keys=True)
        bytes_content = json_content.encode("utf-8")

        self.put_object(bucket=bucket, object_key=object_key, content=bytes_content)



import pandas as pd

from typing import Dict, Union, Any

from google.cloud import bigquery as bq
from google.oauth2 import service_account
from google.auth.credentials import Credentials
from google.cloud.bigquery.job import QueryJob
from google.cloud.bigquery.table import RowIterator
from google.cloud.bigquery.retry import DEFAULT_RETRY
from google.api_core.retry import Retry

from santoku.aws import SecretsManagerHandler


class BigQueryHandler:
    """
    Manage Google BigQuery interactions, the simplest of which is to query a particular table.
    This makes use of the official BigQuery API. More information at
    https://googleapis.dev/python/bigquery/latest/index.html

    """

    def __init__(self, **kwargs) -> None:
        """
        Base constructor. Use this to authenticate using Application Default Credentials.
        For alternative methods of authentication use the given class methods.

        Parameters
        ----------
        kwargs
            Additional arguments to be passed to google.cloud.bigquery.client.Client. If an argument
            is necessary but not provided, Google's API tries to infer it from Application Default
            Credentials.

        Raises
        ------
        google.auth.exceptions.DefaultCredentialsError
            Raised if credentials are not specified and the library fails to acquire default
            credentials.

        Notes
        -----
        Documentation for google.cloud.bigquery.client.Client in [1].
        More on authentication methods in [2] and [3].
        More on Application Default Credentials in [4].
        More on authentication via service account in [5].
        More on authentication as an end user in [6].
        To create a Credentials object, follow the google-auth guide in [7].

        References
        ----------
        [1] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.client.Client.html
        [2] :
        https://cloud.google.com/bigquery/docs/authentication
        [3] :
        https://googleapis.dev/python/google-api-core/latest/auth.html#client-provided-authentication
        [4] :
        https://cloud.google.com/docs/authentication/production
        [5] :
        https://cloud.google.com/docs/authentication/production#obtaining_and_providing_service_account_credentials_manually
        [6] :
        https://cloud.google.com/docs/authentication/end-user
        [7] :
        https://google-auth.readthedocs.io/en/latest/user-guide.html#service-account-private-key-files
        """
        self.client = bq.Client(**kwargs)

    @classmethod
    def from_service_account_info(cls, credential_info: Dict[str, str]):
        """
        Authenticate via service account credentials, parsed as a dictionary.

        Parameters
        ----------
        credential_info: Dict[str, str]
            Credentials for a Google Cloud service account in dictionary form.

        Notes
        -----
        The credentials JSON must look like this:
        ```
        {
            "type": "service_account",
            "project_id": "<project name>",
            "private_key_id": "<private key id>",
            "private_key": "<private key>",
            "client_email": "<service account email>",
            "client_id": "<client id>",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://accounts.google.com/o/oauth2/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/<email>"
        }
        ```
        See [1] for instructions on how to generate such file.

        References
        ----------
        [1] :
        https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating_service_account_keys
        """
        credentials = service_account.Credentials.from_service_account_info(info=credential_info)
        return cls(project=credential_info["project_id"], credentials=credentials)

    @classmethod
    def from_aws_secrets_manager(cls, secret_name: str):
        """
        Retrieve the necessary information to generate service account credentials from AWS Secrets
        Manager. Requires that AWS credentials with the appropriate permissions are located
        somewhere on the AWS credential chain in the local machine.

        Paramaters
        ----------
        secret_name : str
            Name or ARN for the secret containing the JSON needed for BigQuery authentication.

        Notes
        -----
        The retrieved secret must have the particular structure of the JSON file generated by the
        Google Cloud Console, which is associated with a service account. It looks like this:
        ```
        {
            "type": "service_account",
            "project_id": "<project name>",
            "private_key_id": "<private key id>",
            "private_key": "<private key>",
            "client_email": "<service account email>",
            "client_id": "<client id>",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://accounts.google.com/o/oauth2/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/<email>"
        }
        ```
        See [1] for instructions on how to generate such file.

        References
        ----------
        [1] :
        https://cloud.google.com/iam/docs/creating-managing-service-account-keys#creating_service_account_keys
        """
        secrets_manager = SecretsManagerHandler()
        credential_info = secrets_manager.get_secret_value(secret_name=secret_name)
        return cls.from_service_account_info(credential_info=credential_info)

    def _submit_query_job(self, query: str, **kwargs) -> QueryJob:
        """
        Submit a SQL query job against datasets reachable from `self.project` (although that can be
        changed using additional arguments). Returns a native datatype QueryJob, which can be used
        to start the job, wait for it to finish or get the query results.

        Parameters
        ----------
        query : str
            SQL query to be executed. By default on the standard SQL dialect.
        kwargs : Dict[str, Any]
            Additional arguments for google.cloud.bigquery.client.Client.query

        Notes
        -----
        More on additional arguments for the query method: [1]. Note that specific arguments are
        needed if using BigQuery's legacy SQL dialect.
        More information on the QueryJob data type: [2].

        References
        ----------
        [1] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.client.Client.html#google.cloud.bigquery.client.Client.query
        [2] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.job.QueryJob.html

        Returns
        -------
        google.cloud.bigquery.job.QueryJob
            Native datatype representing a BigQuery job. More information in [2].
        """
        return self.client.query(query=query, **kwargs)

    def get_query_results(
        self,
        query: str,
        page_size: int = None,
        max_results: int = None,
        retry: Retry = DEFAULT_RETRY,
        timeout: float = None,
        **kwargs,
    ) -> RowIterator:
        """
        Run a SQL query and return the results as google.cloud.bigquery.table.RowIterator, which is
        BigQuery's native type iterating through API row list responses. This object can be easily
        converted to common datatypes such as pandas dataframes or pyarrow tables (see [3]).

        Parameters
        ----------
        query : str
        page_size : int
        max_results : int
        retry : google.api_core.retry.Retry
        timeout : float

        Notes
        -----
        More on additional arguments for the query method: [1]. Note that specific arguments are
        needed if using BigQuery's legacy SQL dialect.
        More on the google.cloud.bigquery.job.QueryJob.result() method: [2].
        More on the google.cloud.bigquery.table.RowIterator type: [3]
        More on the google.cloud.bigquery.table.Row type: [4].
        More on retries: [5].

        References
        ----------
        [1] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.client.Client.html#google.cloud.bigquery.client.Client.query
        [2] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.job.QueryJob.html#google.cloud.bigquery.job.QueryJob.result
        [3] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.table.RowIterator.html
        [4] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.table.Row.html
        [5] :
        https://googleapis.dev/python/bigquery/latest/reference.html#retries

        Return
        ------
        google.cloud.bigquery.table.RowIterator
            Iterator of google.cloud.bigquery.table.Row containing the results of the query. More
            information in [3].
        """
        return self._submit_query_job(query=query, **kwargs).result(
            page_size=page_size, max_results=max_results, retry=retry, timeout=timeout
        )

    def get_query_results_as_dataframe(
        self,
        query: str,
        page_size: int = None,
        max_results: int = None,
        retry: Retry = DEFAULT_RETRY,
        timeout: float = None,
        dtypes: Dict[str, Union[str, Any]] = None,
        progress_bar_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Run a SQL query and return the results as a pandas dataframe.

        Parameters
        ----------
        query : str
            SQL query to be executed. Using the standard SQL dialect by default.
        page_size: int
            Argument for `self.get_query_results`
        max_results: int
            Argument for `self.get_query_results`
        retry: Retry
            Argument for `self.get_query_results`
        timeout: float
            Argument for `self.get_query_results`
        dtypes: Dict[str, Union[str, Any]]
            A dictionary of column names to pandas dtypes used when constructing the series for the
            column specified. Otherwise, the default pandas behavior is used.
        progress_bar_type: str
            If set, use the tqdm library to display a progress bar while the data downloads.
            Install the tqdm package to use this feature.
        kwargs : Dict[str, Any]
            Additional arguments for google.cloud.bigquery.client.Client.query

        Notes
        -----
        More on additional arguments for the query method: [1]. Note that specific arguments are
        needed if using the BigQuery's legacy SQL flavor.
        More on the to_dataframe() method: [2].

        References
        ----------
        [1] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.client.Client.html#google.cloud.bigquery.client.Client.query
        [2] :
        https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.table.RowIterator.html#google.cloud.bigquery.table.RowIterator.to_dataframe

        Returns
        -------
        pd.DataFrame
            The query results as a pandas dataframe
        """
        return self.get_query_results(
            query=query,
            page_size=page_size,
            max_results=max_results,
            retry=retry,
            timeout=timeout,
            **kwargs,
        ).to_dataframe(dtypes=dtypes, progress_bar_type=progress_bar_type)
